# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Enable consuming Azure open datasets into dataframes and enrich customer data."""

from ._open_image import OpenImage

__all__ = [
    'OpenImage',
]
